diff -u --recursive --new-file --exclude=zImage --exclude=config.ini --exclude=.* --exclude=*.patch --exclude=buildit --exclude=defconfig linux/drivers/block/ide-disk.c hijack/drivers/block/ide-disk.c
--- linux/drivers/block/ide-disk.c	Sat Mar 18 10:23:37 2000
+++ hijack/drivers/block/ide-disk.c	Thu Jun  6 19:34:05 2002
@@ -86,10 +86,23 @@
  * Returns:	1 if lba_capacity looks sensible
  *		0 otherwise
  */
-static int lba_capacity_is_ok (struct hd_driveid *id)
+static int lba_capacity_is_ok (ide_drive_t *drive)
 {
+	struct hd_driveid *id = drive->id;
 	unsigned long lba_sects, chs_sects, head, tail;
+	unsigned short *idw = (unsigned short *)id; 
 
+	if ((idw[83] & 0xc400) == 0x4400) { 		// Does drive support 48-bit LBA?
+		unsigned long capacity;
+		drive->lba28_limit = (idw[61]<<16) | idw[60];	// limit == "normal" lba_capacity value
+		capacity = (idw[101]<<16) | idw[100];
+		// The rest of the kernel can only support 32-bit LBA, so round down:
+		if (idw[103] | idw[102])
+			capacity = ~0;			// 32-bit kernel limit of 2TB
+		id->lba_capacity = capacity;
+		return 1;
+	}
+	drive->lba28_limit = ~0;			// no limit unless 48-bit is supported
 	/*
 	 * The ATA spec tells large drives to return
 	 * C/H/S = 16383/16/63 independent of their size.
@@ -383,6 +398,7 @@
 	ide_hwif_t *hwif = HWIF(drive);
 	int use_pdc4030_io = 0;
 #endif /* CONFIG_BLK_DEV_PDC4030 */
+	int lba48 = 0;
 
 	if (IDE_CONTROL_REG)
 		OUT_BYTE(drive->ctl,IDE_CONTROL_REG);
@@ -402,6 +418,12 @@
 			drive->name, (rq->cmd==READ)?"read":"writ",
 			block, rq->nr_sectors, (unsigned long) rq->buffer);
 #endif
+		if ((block + rq->nr_sectors - 1) > drive->lba28_limit) {	// need to use 48-bit LBA?
+			OUT_BYTE(block>>24,IDE_SECTOR_REG);
+			OUT_BYTE(0,IDE_LCYL_REG);
+			OUT_BYTE(0,IDE_HCYL_REG);
+			lba48 = 1;
+		}
 		OUT_BYTE(block,IDE_SECTOR_REG);
 		OUT_BYTE(block>>=8,IDE_LCYL_REG);
 		OUT_BYTE(block>>=8,IDE_HCYL_REG);
@@ -434,7 +456,10 @@
 			return ide_started;
 #endif /* CONFIG_BLK_DEV_IDEDMA */
 		ide_set_handler(drive, &read_intr, WAIT_CMD);
-		OUT_BYTE(drive->mult_count ? WIN_MULTREAD : WIN_READ, IDE_COMMAND_REG);
+		if (lba48)
+			OUT_BYTE(drive->mult_count ? WIN_MULTREAD_EXT : WIN_READ_EXT, IDE_COMMAND_REG);
+		else
+			OUT_BYTE(drive->mult_count ? WIN_MULTREAD : WIN_READ, IDE_COMMAND_REG);
 		return ide_started;
 	}
 	if (rq->cmd == WRITE) {
@@ -443,7 +468,10 @@
 		if (drive->using_dma && !(HWIF(drive)->dmaproc(ide_dma_write, drive)))
 			return ide_started;
 #endif /* CONFIG_BLK_DEV_IDEDMA */
-		OUT_BYTE(drive->mult_count ? WIN_MULTWRITE : WIN_WRITE, IDE_COMMAND_REG);
+		if (lba48)
+			OUT_BYTE(drive->mult_count ? WIN_MULTWRITE_EXT : WIN_WRITE_EXT, IDE_COMMAND_REG);
+		else
+			OUT_BYTE(drive->mult_count ? WIN_MULTWRITE : WIN_WRITE, IDE_COMMAND_REG);
 		if (ide_wait_stat(&startstop, drive, DATA_READY, drive->bad_wstat, WAIT_DRQ)) {
 			printk(KERN_ERR "%s: no DRQ after issuing %s\n", drive->name,
 				drive->mult_count ? "MULTWRITE" : "WRITE");
@@ -518,14 +546,14 @@
  * current_capacity() returns the capacity (in sectors) of a drive
  * according to its current geometry/LBA settings.
  */
-static unsigned long idedisk_capacity (ide_drive_t  *drive)
+unsigned long idedisk_capacity (ide_drive_t  *drive)
 {
 	struct hd_driveid *id = drive->id;
 	unsigned long capacity = drive->cyl * drive->head * drive->sect;
 
 	drive->select.b.lba = 0;
 	/* Determine capacity, and use LBA if the drive properly supports it */
-	if (id != NULL && (id->capability & 2) && lba_capacity_is_ok(id)) {
+	if (id != NULL && (id->capability & 2) && lba_capacity_is_ok(drive)) {
 		if (id->lba_capacity >= capacity) {
 			capacity = id->lba_capacity;
 			drive->select.b.lba = 1;
diff -u --recursive --new-file --exclude=zImage --exclude=config.ini --exclude=.* --exclude=*.patch --exclude=buildit --exclude=defconfig linux/drivers/block/ide-dma.c hijack/drivers/block/ide-dma.c
--- linux/drivers/block/ide-dma.c	Sat Mar 18 10:23:37 2000
+++ hijack/drivers/block/ide-dma.c	Thu Jun  6 18:35:54 2002
@@ -357,8 +357,10 @@
 			return 0;
 		case ide_dma_check:
 			return config_drive_for_dma (drive);
+		case ide_dma_read48:
 		case ide_dma_read:
 			reading = 1 << 3;
+		case ide_dma_write48:
 		case ide_dma_write:
 			if (!(count = ide_build_dmatable(drive, func)))
 				return 1;	/* try PIO instead of DMA */
@@ -369,7 +371,10 @@
 			if (drive->media != ide_disk)
 				return 0;
 			ide_set_handler(drive, &ide_dma_intr, WAIT_CMD);/* issue cmd to drive */
-			OUT_BYTE(reading ? WIN_READDMA : WIN_WRITEDMA, IDE_COMMAND_REG);
+			if (func == ide_dma_read || func == ide_dma_write48)
+				OUT_BYTE(reading ? WIN_READDMA : WIN_WRITEDMA, IDE_COMMAND_REG);
+			else /* lba48 */
+				OUT_BYTE(reading ? WIN_READDMA_EXT : WIN_WRITEDMA_EXT, IDE_COMMAND_REG);
 		case ide_dma_begin:
 			/* Note that this is done *after* the cmd has
 			 * been issued to the drive, as per the BM-IDE spec.
diff -u --recursive --new-file --exclude=zImage --exclude=config.ini --exclude=.* --exclude=*.patch --exclude=buildit --exclude=defconfig linux/drivers/block/ide.h hijack/drivers/block/ide.h
--- linux/drivers/block/ide.h	Sat Mar 18 10:23:37 2000
+++ hijack/drivers/block/ide.h	Thu Jun  6 19:27:40 2002
@@ -274,6 +287,7 @@
 	byte		sect;		/* "real" sectors per track */
 	byte		bios_head;	/* BIOS/fdisk/LILO number of heads */
 	byte		bios_sect;	/* BIOS/fdisk/LILO sectors per track */
+	unsigned int	lba28_limit;	/* used to detect need for lba48 EXT commands */
 	unsigned short	bios_cyl;	/* BIOS/fdisk/LILO number of cyls */
 	unsigned short	cyl;		/* "real" number of cyls */
 	unsigned int	drive_data;	/* for use by tuneproc/selectproc as needed */
@@ -302,7 +316,7 @@
  */
 typedef enum {	ide_dma_read,	ide_dma_write,	ide_dma_begin,	ide_dma_end,
 		ide_dma_check,	ide_dma_on,	ide_dma_off,	ide_dma_off_quietly,
-		ide_dma_test_irq
+		ide_dma_test_irq, ide_dma_read48, ide_dma_write48
 	} ide_dma_action_t;
 
 typedef int (ide_dmaproc_t)(ide_dma_action_t, ide_drive_t *);
diff -u --recursive --new-file --exclude=zImage --exclude=config.ini --exclude=.* --exclude=*.patch --exclude=buildit --exclude=defconfig linux/drivers/block/ns87415.c hijack/drivers/block/ns87415.c
--- linux/drivers/block/ns87415.c	Sat Mar 18 10:23:37 2000
+++ hijack/drivers/block/ns87415.c	Thu Jun  6 18:37:04 2002
@@ -90,7 +90,9 @@
 			outb(inb(hwif->dma_base)&~1, hwif->dma_base);	/* stop DMA */
 			outb(inb(hwif->dma_base)|6, hwif->dma_base);	/* from ERRATA: clear the INTR & ERROR bits */
 			return (dma_stat & 7) != 4;		/* verify good DMA status */
+		case ide_dma_write48:
 		case ide_dma_write:
+		case ide_dma_read48:
 		case ide_dma_read:
 			ns87415_prepare_drive(drive, 1);	/* select DMA xfer */
 			if (!ide_dmaproc(func, drive))		/* use standard DMA stuff */
diff -u --recursive --new-file --exclude=zImage --exclude=config.ini --exclude=.* --exclude=*.patch --exclude=buildit --exclude=defconfig linux/drivers/block/trm290.c hijack/drivers/block/trm290.c
--- linux/drivers/block/trm290.c	Wed Jun  9 08:09:27 1999
+++ hijack/drivers/block/trm290.c	Thu Jun  6 18:37:54 2002
@@ -178,12 +178,14 @@
 	unsigned int count, reading = 2, writing = 0;
 
 	switch (func) {
+		case ide_dma_write48:
 		case ide_dma_write:
 			reading = 0;
 			writing = 1;
 #ifdef TRM290_NO_DMA_WRITES
 			break;	/* always use PIO for writes */
 #endif
+		case ide_dma_read48:
 		case ide_dma_read:
 			if (!(count = ide_build_dmatable(drive)))
 				break;		/* try PIO instead of DMA */
@@ -194,7 +196,10 @@
 			if (drive->media != ide_disk)
 				return 0;
 			ide_set_handler(drive, &ide_dma_intr, WAIT_CMD);
-			OUT_BYTE(reading ? WIN_READDMA : WIN_WRITEDMA, IDE_COMMAND_REG);
+			if (func == ide_dma_read || func == ide_dma_write)
+				OUT_BYTE(reading ? WIN_READDMA : WIN_WRITEDMA, IDE_COMMAND_REG);
+			else
+				OUT_BYTE(reading ? WIN_READDMA_EXT : WIN_WRITEDMA_EXT, IDE_COMMAND_REG);
 			return 0;
 		case ide_dma_begin:
 			return 0;
diff -u --recursive --new-file --exclude=zImage --exclude=config.ini --exclude=.* --exclude=*.patch --exclude=buildit --exclude=defconfig linux/include/linux/hdreg.h hijack/include/linux/hdreg.h
--- linux/include/linux/hdreg.h	Wed Jun  9 08:09:03 1999
+++ hijack/include/linux/hdreg.h	Thu Jun  6 19:18:12 2002
@@ -81,6 +81,14 @@
 #define WIN_SRST		0x08	/* ATAPI soft reset command */
 #define WIN_PACKETCMD		0xa0	/* Send a packet command. */
 
+/* 48-bit LBA commands */
+#define WIN_READ_EXT		0x24
+#define WIN_READDMA_EXT		0x25
+#define WIN_MULTREAD_EXT	0x29
+#define WIN_WRITE_EXT		0x34
+#define WIN_WRITEDMA_EXT	0x35
+#define WIN_MULTWRITE_EXT	0x39
+
 /* WIN_SMART sub-commands */
 
 #define SMART_READ_VALUES	0xd0
